#include "fuser.h"
#include "process-util.h"

static int parse_dir(struct name *this_name, struct inode *match_inode) {
        if ((this_name == NULL) || (match_inode == NULL)) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't parse dir.");
                return -1;
        }

        if (stat(this_name->filename, &this_name->st) != 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't stat dir %s.", this_name->filename);
                return -1;
        }

        match_inode->name = this_name;
        match_inode->device = this_name->st.st_dev;
        match_inode->inode = this_name->st.st_ino;

        return 0;
}

static int parse_mounts(struct name *this_name, struct device *match_device) {
        if ((this_name == NULL) && (match_device == NULL)) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't parse mounts.");
                return -1;
        }

        match_device->name = this_name;

        if (S_ISBLK(this_name->st.st_mode))
                match_device->device = this_name->st.st_rdev;
        else
                match_device->device = this_name->st.st_dev;

        return 0;
}

static uid_t getpiduid(const pid_t pid) {
        char pathname[PATH_MAX];
        struct stat st;
        int r = 0;

        r = snprintf(pathname, sizeof(pathname), "/proc/%d", pid);
        if (r < 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Snprintf run failed in getpiduid.");
                return 0;
        }

        if (stat(pathname, &st) != 0)
                return 0;

        return st.st_uid;
}

static struct stat *get_pidstat(const pid_t pid) {
        char pathname[PATH_MAX];
        struct stat *st = NULL;
        int r = 0;

        st = (struct stat *)malloc(sizeof(struct stat));
        if (st == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Malloc failed in get_pidstat.");
                return NULL;
        }

        r = snprintf(pathname, sizeof(pathname), "/proc/%d/cwd", pid);
        if (r < 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Snprintf run failed in get_pidstat.");
                return NULL;
        }

        if (stat(pathname, st) != 0) {
                free(st);
                return NULL;
        }

        return st;
}

static void add_matched_proc(struct name *name, const pid_t pid, const uid_t uid) {
        struct procs *pptr = NULL;
        struct procs *last_proc = NULL;
        char pathname[PATH_MAX];
        char cmdname[CMD_NAME_LEN + 1];
        char *cptr = NULL;
        int cmdlen = 0;
        FILE *fp = NULL;

        if (name == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Name should not be NULL.");
                return;
        }

        //find out wheather the pid already in pptr->pid
        for (pptr = name->matched_procs; pptr != NULL; pptr = pptr->next) {
                last_proc = pptr;

                if (pptr->pid == pid)
                            return;
        }

        pptr = (struct procs *)malloc(sizeof(struct procs));
        if (pptr == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't malloc in add_matched_proc.");
                return;
        }

        pptr->pid = pid;
        pptr->uid = uid;
        pptr->username = NULL;
        pptr->next = NULL;
        pptr->command = NULL;

        if ((snprintf(pathname, sizeof(pathname), "/proc/%d/stat", pid) > 0) &&
                ((fp = fopen(pathname, "r")) != NULL) && (fscanf(fp, "%*d (%100[^)]", cmdname) == 1)) {
                pptr->command = (char *)malloc(COMM_LEN + 1);

                if (pptr->command != NULL) {
                       cmdlen = 0;

                       for (cptr = cmdname; cmdlen < COMM_LEN && *cptr; cptr++) {
                               if (isprint(*cptr)) {
                                       pptr->command[cmdlen++] = *cptr;
                               } else if (cmdlen < (COMM_LEN - 4)) {
                                       cmdlen += sprintf(&(pptr->command[cmdlen]), "\\%03o", (unsigned int)*cptr);
                               }
                       }

                       pptr->command[cmdlen] = '\0';
                }
        }

        if (last_proc == NULL)
                name->matched_procs = pptr;
        else
                last_proc->next = pptr;

        if (fp)
                fclose(fp);
}

static void check_dir(const pid_t pid, const char *dirname, const struct device *dev,
        const struct inode *ino, const uid_t uid) {
        DIR *dirp = NULL;
        dev_t thedev;
        struct dirent *direntry = NULL;
        struct stat st;
        char dirpath[PATH_MAX];
        char filepath[PATH_MAX];
        int r = 0;

        if (dirname == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Dirname is NULL.");
                return;
        }

        r = snprintf(dirpath, sizeof(dirpath), "/proc/%d/%s", pid, dirname);
        if (r < 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Snprintf run failed in check_dir.");
               return;
        }

        dirp = opendir(dirpath);
        if (dirp == NULL)
                return;

        while ((direntry = readdir(dirp)) != NULL) {
                if (direntry->d_name[0] < '0' || direntry->d_name[0] > '9')
                        continue;

                snprintf(filepath, sizeof(filepath), "/proc/%d/%s/%s",
                        pid, dirname, direntry->d_name);

                if (stat(filepath, &st) != 0)
                        continue;

                thedev = st.st_dev;

                if ((dev != NULL) && (thedev == dev->device)) {
                        add_matched_proc(dev->name, pid, uid);
                }

                if ((ino != NULL) && (thedev == ino->device)) {
                        if (st.st_ino == ino->inode) {
                                add_matched_proc(ino->name, pid, uid);
                        }
                }
        } //end while

        closedir(dirp);
}

static int scan_procs(const struct name *name, const struct inode *ino, const struct device *dev) {
        DIR *topproc_dir = NULL;
        struct dirent *topproc_dent = NULL;
        pid_t pid;
        pid_t my_pid;
        uid_t uid;

        if (name == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                         "Name should not be null in scan_procs.");
                return -1;
        }

        if ((ino == NULL) && (dev == NULL)) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Ino and dev should not be NULL in scan_procs.");
                return -1;
        }

        topproc_dir = opendir("/proc");
        if (topproc_dir == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't open dir proc.");
                return -1;
        }

       my_pid = getpid();

        while ((topproc_dent = readdir(topproc_dir)) != NULL) {
                dev_t scan_dev;
                struct stat *st = NULL;

                /* Not a process */
                if ((topproc_dent->d_name[0] < '0') || (topproc_dent->d_name[0] > '9'))
                        continue;

                pid = atoi(topproc_dent->d_name);
                if (pid == my_pid)
                        continue;

                uid = getpiduid(pid);

                st = get_pidstat(pid);
                scan_dev = st ? st->st_dev : 0;

                if ((dev != NULL) && (scan_dev == dev->device))
                        add_matched_proc(dev->name, pid, uid);

                if ((ino != NULL) && (scan_dev == ino->device)) {
                        if (!st)
                                st = get_pidstat(pid);

                        if (st && (st->st_dev == ino->device) && (st->st_ino == ino->inode))
                                add_matched_proc(ino->name, pid, uid);
                }

                if (st)
                        free(st);

                check_dir(pid, "fd", dev, ino, uid);
        } // end while

        closedir(topproc_dir);
        return 0;
}

static void add_special_proc(struct name *name, const uid_t uid, const char *command) {
        struct procs *pptr = NULL;

        if (name == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Name should not be null in add_special_proc.");
                return;
        }

        for (pptr = name->matched_procs; pptr != NULL; pptr = pptr->next) {
                if (pptr->command != NULL && strcmp(pptr->command, command) == 0)
                        return;
        }

        if ((pptr = malloc(sizeof(struct procs))) == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't allocate memory for add_special_proc() proc");
                return;
        }

        pptr->pid = 0;
        pptr->uid = uid;
        pptr->next = name->matched_procs;
        pptr->command = strdup(command);

        name->matched_procs = pptr;
}

static void scan_mounts_and_swaps(const struct name *name, const struct inode *ino,
            const struct device *dev, const char *file) {
        FILE *fp = NULL;
        char line[PATH_MAX];
        char *find_mountp = NULL;
        char *find_space_mounts = NULL;
        char *find_space_swaps = NULL;
        struct stat st;

        if (name == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Name should not be null in scan_mounts_and_swaps.");
                return;
        }

        if ((ino == NULL) && (dev == NULL)) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Ino and dev should not be null in scan_mounts_and_swaps.");
                return;
        }

        fp = fopen(file, "r");
        if (fp == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't open file %s", file);
                return;
        }

        while (fgets(line, PATH_MAX, fp) != NULL) {
                if (strcmp(file, PROC_MOUNTS) == 0) {
                        if ((find_mountp = strchr(line, ' ')) == NULL)
                                continue;

                        find_mountp++;

                        find_space_mounts = strchr(find_mountp, ' ');
                        if (find_space_mounts == NULL)
                                continue;

                        *find_space_mounts = '\0';

                        if (stat(find_mountp, &st) != 0)
                                continue;
                } else {
                        find_space_swaps = strchr(line, ' ');
                        if (find_space_swaps == NULL)
                                continue;

                        *find_space_swaps = '\0';
                        find_space_swaps++;

                        while (*find_space_swaps == ' ') {
                                find_space_swaps++;

                                if (*find_space_swaps == '\0')
                                        continue;
                        }

                        if (stat(line, &st) != 0) {
                                continue;
                        }
                }

                if ((dev != NULL) && (st.st_dev == dev->device)) {
                        if (strcmp(file, PROC_MOUNTS) == 0)
                                add_special_proc(dev->name, 0, find_mountp);

                        if (strcmp(file, PROC_SWAPS) == 0)
                                add_special_proc(dev->name, 0, line);
                }

                if ((ino != NULL) && (st.st_dev == ino->device) && (st.st_ino == ino->inode)) {
                        if (strcmp(file, PROC_MOUNTS) == 0)
                                add_special_proc(ino->name, 0, find_mountp);

                        if (strcmp(file, PROC_SWAPS) == 0)
                                add_special_proc(ino->name, 0, line);
                }
        } // end while

        fclose(fp);
}

static void print_matches(const struct name *name) {
        struct procs *pptr = NULL;
        struct passwd *pwent = NULL;
        static char P_cmd_long[MAX_COMM_LEN];
        char cmd_path[PATH_MAX];
        int r = 0;
        FILE *fp = NULL;
        int flag = 0;

        if (name == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Name should not be null in print_matches.");
                return;
        }

        /* Write the content in the back of previous one */
        fp = fopen(REBOOT_UMOUNT_FILE_NAME, "a+");

        /* print the time info to /.reboot-umount-msg.log file */
        if (fp == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Open %s failed!", REBOOT_UMOUNT_FILE_NAME);
        }

        manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                "\t\tUSER\t\tPID\tCOMMAND");
        manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                "%s:", name->filename);

        /* print the umount fail point to the /.reboot-umount-msg.log file */
        if (fp != NULL) {
                if (strlen(name->filename) <= MOUNT_FILE_NAME_MAX_LEN) {
                        fprintf(fp, "%-20s\t", name->filename);
                } else {
                        fprintf(fp, "%s\n\t\t\t", name->filename);
                }
        }

        for (pptr = name->matched_procs; pptr != NULL; pptr = pptr->next) {
                if (pwent == NULL || pwent->pw_uid != pptr->uid)
                        pwent = getpwuid(pptr->uid); //get username

                r = snprintf(cmd_path, sizeof(cmd_path), "/proc/%d", pptr->pid);
                if (r <= 0) {
                        manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL, "Can't snprintf /proc/%d.", pptr->pid);
                        goto out;
                }

                read_cmdline(P_cmd_long, sizeof(P_cmd_long), cmd_path, "cmdline", ' ');

                if (strlen(P_cmd_long) != 0){
                        free(pptr->command);
                        pptr->command = strdup(P_cmd_long);
                }

                if (pptr->command == NULL)
                        continue;

                if (flag > 0) {
                        if (fp != NULL) {
                                fprintf(fp, "\t\t\t");
                        }
                } else {
                        flag++;
                }

                if (pwent != NULL) {
                        if (pptr->pid != 0) {
                                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                                        "\t\t%-s\t\t%-d\t%-s", pwent->pw_name, pptr->pid, pptr->command);
                                if (fp != NULL) {
                                        fprintf(fp, "%-s\t\t%-d\t%-s\n", pwent->pw_name, pptr->pid, pptr->command);
                                }
                        } else {
                                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                                        "\t\t%-s\t\t%-s\t%-s", pwent->pw_name, "kernel", pptr->command);
                                if (fp != NULL) {
                                        fprintf(fp, "%-s\t\t%-s\t%-s\n", pwent->pw_name, "kernel", pptr->command);
                                }
                        }
                } else {
                        if (pptr->pid != 0) {
                                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                                        "\t\t%-u\t\t%-d\t%-s", pptr->uid, pptr->pid, pptr->command);
                                if (fp != NULL) {
                                        fprintf(fp, "%-u\t\t%-d\t%-s\n", pptr->uid, pptr->pid, pptr->command);
                                }
                        } else {
                                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                                      "\t\t%-u\t\t%-s\t%-s", pptr->uid, "kernel", pptr->command);
                                if (fp != NULL) {
                                        fprintf(fp, "%-u\t\t%-s\t%-s\n", pptr->uid, "kernel", pptr->command);
                                }
                        }
                }
        }

out:
        if (fp != NULL) {
                fclose(fp);
        }
}

static void free_matched_procs(struct procs *matched_procs) {
        struct procs *procs_tmp = NULL;
        struct procs *procs_next = NULL;

        procs_tmp = matched_procs;

        while (procs_tmp != NULL) {
                procs_next = procs_tmp->next;

                if (procs_tmp->command)
                        free(procs_tmp->command);

                free(procs_tmp);

                procs_tmp = procs_next;
        }
}

int fuser(const char *dir) {
        struct name this_name;
        struct inode match_inode;
        struct device match_device;
        int r = 0;

        if (dir == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Dir should not be NULL.");
                return -1;
        }

        this_name.matched_procs = NULL;

        this_name.filename = strdup(dir); //need to free
        if (this_name.filename == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                      "Can't allocate memory for fuser() this_name->filename.");
                return -1;
        }

        r = parse_dir(&this_name, &match_inode);
        if (r < 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "%s", "Failed to parse file.");
                free(this_name.filename);
                return -1;
        }

        r = parse_mounts(&this_name, &match_device);
        if (r < 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "%s", "Failed to parse mounts.");
                free(this_name.filename);
                return -1;
        }

        r = scan_procs(&this_name, &match_inode, &match_device);
        if (r < 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "%s", "Failed to scan_procs.");
                free(this_name.filename);
                return -1;
        }

        scan_mounts_and_swaps(&this_name, &match_inode, &match_device, PROC_MOUNTS);
        scan_mounts_and_swaps(&this_name, &match_inode, &match_device, PROC_SWAPS);
        print_matches(&this_name);

        free_matched_procs(this_name.matched_procs);
        free(this_name.filename);
        return 0;
}
